@@ -13,8 +13,9 @@ import subprocess
 import os
 import fnmatch
 import pathlib
+import inspect
 from collections import OrderedDict
-from datetime import datetime, timedelta
+from datetime import datetime, timedelta, date as dt_date
 
 
 # ZFS connection classes
@@ -247,18 +248,20 @@ class ZFSItem(object):
         self._properties = {}
         if parent:
             self.parent = parent
-            self.parent.add_child(self)
+            self.parent._add_child(self)
 
-    def add_child(self, child):
+    def _add_child(self, child):
         self.children.append(child)
         return child
 
+
     def get_child(self, name):
-        child = [ c for c in self.children if c.name == name and isinstance(c, Dataset) ]
+        child = [ c for c in self.children if c.name == name ]
         assert len(child) < 2
         if not child: raise KeyError(name)
         return child[0]
         
+
     def remove(self, child):
         if child not in self.children: raise KeyError(child.name)
         child.invalidated = True
@@ -303,14 +306,14 @@ class Snapable(ZFSItem): # Abstract class for Pools and Datasets
 
         if "/" not in path:
             try: ret = self.get_child(path)
-            except KeyError: raise KeyError("No such dataset %s at %s" % (path, self.path))
+            except KeyError: raise KeyError("No such dataset %s under %s" % (path, self.path))
             if snapshot:
                 try: ret = ret.get_snapshot(snapshot)
-                except KeyError: raise KeyError("No such snapshot %s at %s" % (snapshot, ret.path))
+                except KeyError: raise KeyError("No such snapshot %s under %s" % (snapshot, ret.path))
         else:
             head, tail = path.split("/", 1)
             try: child = self.get_child(head)
-            except KeyError: raise KeyError("No such dataset %s at %s" % (head, self.path))
+            except KeyError: raise KeyError("No such dataset %s under %s" % (head, self.path))
             if snapshot: tail = tail + "@" + snapshot
             ret = child.lookup(tail)
 
@@ -324,17 +327,38 @@ class Snapable(ZFSItem): # Abstract class for Pools and Datasets
     path = property(_get_path)
 
 
+    # For name, use full dataset path
+    def get_dataset(self, name):
+        assert not(isinstance(self, Snapshot)), "get_dataset(name) cannot be used on Snapshot Objects. Use Snapshot.dataset instead."
+        allds = self.get_all_datasets()
+        pool_name = self.name if isinstance(self, Pool) else self.pool.name
+        nfind = name if name.find(pool_name+'/') == 0 else '{}/{}'.format(pool_name, name)
+        for dataset in allds:
+            if dataset.path == nfind:
+                return dataset
+        raise KeyError("Dataset '{}' not found in pool '{}'.".format(name, self.pool.name))
+        
 
-    # if index is True return list(of tuple(int, Snapshot, dataset))
+    # returns list(of str) or if with_depth == True then list(of tuple(of depth, Dataset))
+    def get_all_datasets(self, with_depth:bool=False, depth:int=0):
+        a = []
+        for c in self.children:
+            if isinstance(c, Dataset):
+                a.append( (depth, c) if with_depth else c )
+                a = a + c.get_all_datasets(with_depth=with_depth, depth=depth+1)
+        return a
+
+
+    # if index is True return list(of tuple(int, Snapshot))
     def get_snapshots(self, flt=True, index=False):
         if flt is True: flt = lambda _:True
-        if flt is None:
-            assert 0, "flt must not be None"
+        assert inspect.isfunction(flt), "flt must either be True or a Function. Got: {}".format(type(flt))
+        assert isinstance(index, bool), "index must be a boolean. Got: {}".format(type(index))
         _ds_path = self.path
         res = []
         for idx, c in enumerate(self.children):
             if isinstance(c, Snapshot) and flt(c):
-                res.append( (idx, c, self.path) if index else c )
+                res.append( (idx, c) if index else c )
 
         return res
 
@@ -357,30 +381,33 @@ class Snapable(ZFSItem): # Abstract class for Pools and Datasets
     #  - dt_from: datetime to start
     #  - tdelta: timedelta -or- string of nC where: n is an integer > 0 and C is one of y,m,d,H,M,S. Eg 5H = 5 Hours
     #  - dt_to: datetime to stop 
-    #  - index: (bool) - Return list(tuple(of int, snapshot, dataset)) where int is the index in current snaphot listing for dataset
+    #  - index: (bool) 
+    #  Return: 
+    #  -  list(tuple(of int, snapshot)) where int is the index in current snaphot listing for dataset
     #  Notes:
     #  - Date searching is any combination of:
     #      (dt_from --> dt_to) | (dt_from --> dt_from + tdelta) | (dt_to - tdelta --> dt_to) | (dt_from --> now)
     def find_snapshots(self, find_opts:dict) -> list:
 
-        def __assert(k, types):
+        def __assert(k, types, default=None, to_datetime=False):
             if k == 'find_opts':
                 v=find_opts
             else:
-                if not k in find_opts: return None
+                if not k in find_opts: return default
                 v = find_opts[k]
-            bOk=False
-            for t in types:
-                if isinstance(v, t): bOk=True
-            assert bOk, 'Invalid type for param {}. Expecting {} but got: {}'.format(k, types, type(v))
+            assert isinstance(v, types), 'Invalid type for param {}. Expecting {} but got: {}'.format(k, types, type(v))
+
+            if to_datetime and not isinstance(v, datetime):
+                return datetime(v.year, v.month, v.day)
             return v
-        find_opts = __assert('find_opts', [dict])
-        name = __assert('name', [str])
-        dt_from = __assert('dt_from', [datetime])
-        dt_to = __assert('dt_to', [datetime])
-        tdelta = __assert('tdelta', [str, timedelta])
-        index = __assert('index', [bool])
-        contains = __assert('contains', [str])
+
+        find_opts = {} if find_opts is None else __assert('find_opts', (dict))
+        name = __assert('name', (str))
+        dt_from = __assert('dt_from', (datetime, dt_date), None, True)
+        dt_to = __assert('dt_to', (datetime, dt_date), None, True)
+        tdelta = __assert('tdelta', (str, timedelta))
+        index = __assert('index', (bool), False)
+        contains = __assert('contains', (str))
 
         if not contains is None:
             if not os.path.exists(contains):
@@ -410,7 +437,16 @@ class Snapable(ZFSItem): # Abstract class for Pools and Datasets
             f=__fil_dt
 
         elif not tdelta is None and dt_from is None and dt_to is None:
-            (dt_f, dt_t) = (datetime.now() - buildTimedelta(tdelta), datetime.now())
+            tdelta = tdelta if isinstance(tdelta, timedelta) else buildTimedelta(tdelta)
+            (dt_f, dt_t) = (datetime.now() - tdelta, datetime.now())
+            f=__fil_dt
+
+        elif not dt_from is None and not dt_to is None:
+            if not tdelta is None:
+                raise KeyError("tdelta cannot be specified when both dt_from and dt_to are specified")
+            if dt_from >= dt_to:
+                raise KeyError("dt_from ({}) must be < dt_to ({})".format(dt_from, dt_to))
+            (dt_f, dt_t) = (dt_from, dt_to)
             f=__fil_dt
 
         else:
@@ -428,26 +464,7 @@ class Snapable(ZFSItem): # Abstract class for Pools and Datasets
 
 
 
-    # For name, use full dataset path
-    def get_dataset(self, name):
-        assert not(isinstance(self, Snapshot)), "get_dataset(name) cannot be used on Snapshot Objects. Use Snapshot.dataset instead."
-        allds = self.get_all_datasets()
-        pool_name = self.name if isinstance(self, Pool) else self.pool.name
-        nfind = name if name.find(pool_name+'/') == 0 else '{}/{}'.format(pool_name, name)
-        for dataset in allds:
-            if dataset.path == nfind:
-                return dataset
-        raise ValueError("Dataset '{}' not found in pool '{}'.".format(name, self.pool.name))
-        
 
-    # returns list(of str) or if with_depth == True then list(of tuple(of depth, Dataset))
-    def get_all_datasets(self, with_depth:bool=False, depth:int=0):
-        a = []
-        for c in self.children:
-            if isinstance(c, Dataset):
-                a.append( (depth, c) if with_depth else c )
-                a = a + c.get_all_datasets(with_depth=with_depth, depth=depth+1)
-        return a
 
 
 class Pool(Snapable):
@@ -455,6 +472,7 @@ class Pool(Snapable):
         super().__init__(self, name)
         self.connection = conn
         self.have_mounts = have_mounts
+        self.pool = self
 
         
     def __str__(self):
@@ -466,10 +484,13 @@ class Pool(Snapable):
 
 class Dataset(Snapable):
 
+    dspath=None
+    _mountpoint=None
+    _mounted=None
+    
     def __init__(self, pool, name, parent=None):
         super().__init__(pool, name, parent)
         self.dspath = self.path[len(pool.name)+1:]
-        self._mountpoint=None
 
 
     # get_diffs() - Gets Diffs in snapshot or between snapshots (if snap_to is specified)
@@ -562,6 +583,16 @@ class Dataset(Snapable):
         return self._mountpoint
     mountpoint = property(_get_mountpoint)
 
+
+    def _get_mounted(self):
+        if self._mounted is None:
+            self.assertHaveMounts()
+            self._mounted = True if self.get_property('mounted') == 'yes' else False
+        return self._mounted
+    mounted = property(_get_mounted)
+
+    has_mount = property(lambda self: False if self.mountpoint == 'none' else True)
+
     # def _get_path(self):
     #     if not self.parent: return self.name
     #     return "%s/%s" % (self.parent.path, self.name)
@@ -583,12 +614,6 @@ class Dataset(Snapable):
         assert self.pool.have_mounts, "Mount information not loaded. Please use Connection.load_poolset(get_mounts=True)."
 
 
-
-    def get_relative_name(self):
-        if not self.parent: return self.name
-        return self.path[len(self.parent.path) + 1:]
-
-
     def __str__(self):
         if self.pool.have_mounts:
             return "<Dataset:  %s> mountpoint: %s" % (self.path, self.mountpoint)
@@ -821,12 +846,7 @@ def calcDateRange(tdelta:str, dt_from:datetime=None, dt_to:datetime=None) -> tup
     elif dt_to and not isinstance(dt_to, datetime):
         raise KeyError('dt_to must be  a datetime')
 
-
-    if isinstance(tdelta, timedelta):
-        td = tdelta
-        
-    else:
-        td = buildTimedelta(tdelta)
+    td = tdelta if isinstance(tdelta, timedelta) else buildTimedelta(tdelta)
     
     if dt_from:
         return (dt_from, (dt_from + td))
